ACT_ARIO_SPIN = allocate_mario_action(ACT_GROUP_MOVING | ACT_FLAG_MOVING | ACT_FLAG_ATTACKING)
ACT_ARIO_AIR_SPIN = allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_AIR | ACT_FLAG_ATTACKING |
    ACT_FLAG_ALLOW_VERTICAL_WIND_ACTION)
ACT_ARIO_WATER_SPIN = allocate_mario_action(ACT_GROUP_SUBMERGED | ACT_FLAG_MOVING | ACT_FLAG_SWIMMING |
    ACT_FLAG_SWIMMING_OR_FLYING | ACT_FLAG_WATER_OR_TEXT | ACT_FLAG_ATTACKING)
ACT_ARIO_CAP_THROW = allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_MOVING | ACT_FLAG_AIR)
ACT_ARIO_ROLL = allocate_mario_action(ACT_GROUP_MOVING | ACT_FLAG_MOVING | ACT_FLAG_ATTACKING)
ACT_ARIO_WATER_ROLL = allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_MOVING | ACT_FLAG_ATTACKING)
ACT_ARIO_AIR_ROLL = allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_AIR | ACT_FLAG_ATTACKING |
    ACT_FLAG_ALLOW_VERTICAL_WIND_ACTION)
ACT_ARIO_CAP_HIDE = allocate_mario_action(ACT_GROUP_OBJECT | ACT_FLAG_MOVING | ACT_FLAG_SHORT_HITBOX)
ACT_ARIO_CAP_HIDE_WATER = allocate_mario_action(ACT_GROUP_SUBMERGED | ACT_FLAG_SHORT_HITBOX)

local SFX_STAR = audio_sample_load('sfx_star.ogg')

---@param m MarioState
local function ario_spin_loop(m)
    if m.actionTimer == 0 then
        m.forwardVel = math.max(math.min(m.forwardVel + 10.0, 50.0), m.forwardVel)
        set_mario_animation(m, CHAR_ANIM_TWIRL)
        audio_sample_play(SFX_STAR, m.pos, 1.0)
    end
    smlua_anim_util_set_animation(m.marioObj, 'ario_spin_ground')

    local angleDiff = abs_angle_diff(m.intendedYaw, m.faceAngle.y)
    if m.input & INPUT_NONZERO_ANALOG ~= 0 then
        m.forwardVel = math.max(5.0, m.forwardVel - 1.0 * (angleDiff / 0x8000))
    else
        m.forwardVel = math.max(5.0, m.forwardVel - 1.0)
    end

    if angleDiff < 0x6000 then
        m.faceAngle.y = approach_s16_symmetric(m.faceAngle.y, m.intendedYaw, 0x280 + m.actionTimer * 16)
    end

    mario_set_forward_vel(m, m.forwardVel)
    if abs_angle_diff(m.floorAngle, m.faceAngle.y) > 0x4000 then
        local surface = mario_get_floor_class(m)
        if surface == SURFACE_CLASS_VERY_SLIPPERY then
            m.forwardVel = approach_f32_symmetric(m.forwardVel, 0.0, 3.0)
        elseif surface == SURFACE_CLASS_SLIPPERY then
            m.forwardVel = approach_f32_symmetric(m.forwardVel, 0.0, 1.0)
        end
    end

    local step = perform_ground_step(m)

    if m.actionTimer >= 30 then
        return set_mario_action(m, ACT_WALKING, 0)
    end
    if m.actionTimer > 3 then
        if m.controller.buttonPressed & A_BUTTON ~= 0 then
            return set_mario_action(m, ACT_DOUBLE_JUMP, 0)
        end
        if m.controller.buttonPressed & X_BUTTON ~= 0 and not gPlayerSyncTable[m.playerIndex].capThrown and m.flags & MARIO_NORMAL_CAP ~= 0 then
            return set_mario_action(m, ACT_ARIO_CAP_THROW, 0)
        end
    end
    --[[
    if m.actionTimer > 16 then
        if m.controller.buttonPressed & Y_BUTTON ~= 0 then
            m.actionTimer = 0
            set_anim_to_frame(m, 0)
            return 0
        end
    end
    ]]
    if step == GROUND_STEP_HIT_WALL then
        return slide_bonk(m, ACT_BACKWARD_GROUND_KB, ACT_BRAKING_STOP)
    elseif step == GROUND_STEP_LEFT_GROUND then
        return set_mario_action(m, ACT_FREEFALL, 0)
    end

    m.marioObj.header.gfx.angle.x = 0x0500

    if m.actionTimer <= 20 then
        m.marioBodyState.handState = 1
        set_mario_particle_flags(m, PARTICLE_SPARKLES, 0)
    end

    if m.actionTimer == 1 or m.actionTimer == 6 then
        play_sound(SOUND_GENERAL_SWISH_WATER, m.marioObj.header.gfx.cameraToObject)
    end

    m.actionTimer = m.actionTimer + 1
end

---@param m MarioState
local function ario_air_spin_loop(m)
    if m.actionTimer == 0 then
        m.vel.y = 25.0 * (1.0 + m.numStars / 120)
        set_mario_animation(m, CHAR_ANIM_TWIRL)
        audio_sample_play(SFX_STAR, m.pos, 1.0)
    end
    smlua_anim_util_set_animation(m.marioObj, 'ario_spin_air')

    local step = common_air_action_step(m, ACT_FREEFALL_LAND, CHAR_ANIM_TWIRL, AIR_STEP_CHECK_LEDGE_GRAB)

    if m.actionTimer >= 20 then
        return set_mario_action(m, ACT_FREEFALL, 0)
    end
    if m.controller.buttonPressed & X_BUTTON ~= 0 and m.playerIndex == 0 and m.actionTimer > 3 and not gPlayerSyncTable[0].capThrown and m.flags & MARIO_NORMAL_CAP ~= 0 then
        if _G.charSelect.character_get_current_costume(0) == 1 then
            if not gStateExtras[0].restrictCapThrow then
                gPlayerSyncTable[0].capThrown = true
                ---@param o Object
                spawn_sync_object(id_bhvArioThrownCap, CAPMODELS_ARIO_CLOUD.normal, m.pos.x, m.pos.y + 20, m.pos.z,
                    function(o)
                        o.globalPlayerIndex = m.marioObj.globalPlayerIndex
                        o.oMoveAngleYaw = m.faceAngle.y
                        o.oVelY = -40.0
                        o.oAction = 0
                        -- skip 3 frames of the timer so the platform appears quicker
                        o.oTimer = 3
                        o.oBehParams = 1
                        play_sound(SOUND_ACTION_THROW, o.header.gfx.cameraToObject)
                    end)
                m.vel.y = 8.0
                m.forwardVel = 0.0
            end
        else
            return set_mario_action(m, ACT_ARIO_CAP_THROW, 1)
        end
    end

    if m.actionTimer <= 12 then
        m.marioBodyState.handState = 1
        set_mario_particle_flags(m, PARTICLE_SPARKLES, 0)
    end

    if m.actionTimer == 1 or m.actionTimer == 5 then
        play_sound(SOUND_GENERAL_SWISH_WATER, m.marioObj.header.gfx.cameraToObject)
    end

    m.actionTimer = m.actionTimer + 1
end

---@param m MarioState
local function ario_air_spin_gravity(m)
    m.vel.y = m.vel.y - 3.0
end

---@param m MarioState
local function ario_water_spin_loop(m)
    if m.actionTimer == 0 then
        m.forwardVel = 30.0 * (1.0 + m.numStars / 240)
        set_mario_animation(m, CHAR_ANIM_TWIRL)
        audio_sample_play(SFX_STAR, m.pos, 1.0)
    end
    smlua_anim_util_set_animation(m.marioObj, 'ario_spin_water')

    update_flying(m)
    perform_water_step(m)

    if m.actionTimer >= 30 then
        return set_mario_action(m, ACT_WATER_IDLE, 0)
    end

    if m.actionTimer <= 16 then
        m.marioBodyState.handState = 1
        set_mario_particle_flags(m, PARTICLE_SPARKLES, 0)

        if m.pos.y >= m.waterLevel - 80 and m.faceAngle.x > 0x1800 then
            m.pos.y = m.waterLevel
            local vel = m.forwardVel * 1.5
            m.vel.y = sins(m.faceAngle.x) * vel
            m.forwardVel = coss(m.faceAngle.x) * vel
            return set_mario_action(m, ACT_DIVE, 0)
        end
    end

    if m.actionTimer == 1 or m.actionTimer == 5 then
        play_sound(SOUND_GENERAL_SWISH_WATER, m.marioObj.header.gfx.cameraToObject)
    end

    m.actionTimer = m.actionTimer + 1
end

hook_mario_action(ACT_ARIO_SPIN, ario_spin_loop, INT_SLIDE_KICK)
hook_mario_action(ACT_ARIO_AIR_SPIN, { every_frame = ario_air_spin_loop, gravity = ario_air_spin_gravity },
    INT_SLIDE_KICK)
hook_mario_action(ACT_ARIO_WATER_SPIN, ario_water_spin_loop)


---@param m MarioState
local function ario_cap_throw_loop(m)
    set_mario_animation(m, CHAR_ANIM_THROW_LIGHT_OBJECT)
    if m.actionTimer == 2 then
        if m.playerIndex == 0 then
            gPlayerSyncTable[0].capThrown = true
            ---@param o Object
            spawn_sync_object(id_bhvArioThrownCap, CAPMODELS_ARIO.normal, m.pos.x, m.pos.y + 20, m.pos.z, function(o)
                o.globalPlayerIndex = m.marioObj.globalPlayerIndex
                o.oMoveAngleYaw = m.faceAngle.y
                o.oForwardVel = 50.0 + m.forwardVel * 0.8
                o.oAction = 0
                if _G.charSelect.character_get_current_costume(m.playerIndex) == 1 then
                    o.oBehParams = 1
                    obj_set_model_extended(o, CAPMODELS_ARIO_CLOUD.normal)
                end
                play_sound(SOUND_ACTION_THROW, o.header.gfx.cameraToObject)
            end)
        end
    end

    if m.actionArg == 0 then
        -- grounded

        m.forwardVel = approach_f32_symmetric(m.forwardVel, 0.0, 2.0)
        local step = perform_ground_step(m)

        if m.actionTimer > 5 then
            if m.input & INPUT_NONZERO_ANALOG ~= 0 then
                return set_mario_action(m, ACT_WALKING, 0)
            end
            if m.actionTimer > 10 then
                return set_mario_action(m, ACT_WALKING, 0)
            end
        end

        if step == GROUND_STEP_LEFT_GROUND then
            return set_mario_action(m, ACT_FREEFALL, 0)
        end
    else
        --aerial

        update_air_without_turn(m)
        local step = perform_air_step(m, 0)

        if m.actionTimer > 3 then
            if m.controller.buttonPressed & Z_TRIG ~= 0 then
                return set_mario_action(m, ACT_GROUND_POUND, 0)
            elseif m.controller.buttonPressed & B_BUTTON ~= 0 then
                return set_mario_action(m, ACT_DIVE, 0)
            elseif m.controller.buttonPressed & Y_BUTTON ~= 0 and not gStateExtras[m.playerIndex].airSpinUsed then
                return set_mario_action(m, ACT_ARIO_AIR_SPIN, 0)
            end
            if m.actionTimer > 6 then
                return set_mario_action(m, ACT_FREEFALL, 0)
            end
        end
        if step == AIR_STEP_LANDED then
            return set_mario_action(m, ACT_FREEFALL_LAND, 0)
        end
    end
    m.actionTimer = m.actionTimer + 1
end
---@param m MarioState
local function ario_cap_throw_gravity(m)
    if m.actionArg == 1 then
        m.vel.y = m.vel.y - 2.0
    end
end

hook_mario_action(ACT_ARIO_CAP_THROW, { every_frame = ario_cap_throw_loop, gravity = ario_cap_throw_gravity })

---@param m MarioState
local function ario_roll_loop(m)
    local intendedDYaw = m.intendedYaw - m.slideYaw
    local forward = coss(intendedDYaw)
    local sideward = sins(intendedDYaw)

    if m.input & INPUT_NONZERO_ANALOG == 0 then
        forward = 0.5
    end

    local turnFactor = sideward * (1 / maxf(1, m.forwardVel) * 3)
    local oldX = m.slideVelX
    m.slideVelX = m.slideVelX + m.slideVelZ * (m.intendedMag / 32.0) * turnFactor;
    m.slideVelZ = m.slideVelZ - oldX * (m.intendedMag / 32.0) * turnFactor;

    if m.forwardVel < 10.0 then
        m.slideVelX = m.slideVelX + (m.intendedMag / 32.0) * sins(m.intendedYaw)
        m.slideVelZ = m.slideVelZ + (m.intendedMag / 32.0) * coss(m.intendedYaw)
    end
    if m.forwardVel > 0.02 then
        update_sliding_angle(m, 7.0, m.intendedMag / 32 * forward * 0.02 + 0.98)
        m.faceAngle.y = approach_s16_asymptotic(m.faceAngle.y, m.slideYaw, 2)
    else
        m.forwardVel = math.sqrt(sqr(m.slideVelX) + sqr(m.slideVelZ))
    end

    local vel = math.sqrt(sqr(m.slideVelX) + sqr(m.slideVelZ))

    if vel < 8.0 then
        -- slow speed
        if m.controller.buttonDown & Z_TRIG == 0 then
            return set_mario_action(m, ACT_STOP_CROUCHING, 0)
        end
        if mario_floor_is_slippery(m) == 0 then
            if m.controller.buttonPressed & A_BUTTON ~= 0 then
                return set_mario_action(m, ACT_BACKFLIP, 0)
            end
            if m.controller.buttonPressed & B_BUTTON ~= 0 then
                return set_mario_action(m, ACT_PUNCHING, 9)
            end
        end
    else
        -- moving speed
        if m.controller.buttonDown & Z_TRIG == 0 then
            return set_mario_action(m, ACT_BRAKING, 0)
        end
        if mario_floor_is_slippery(m) == 0 or (abs_angle_diff(m.slideYaw, m.floorAngle) < 0x4000) then
            if m.controller.buttonPressed & A_BUTTON ~= 0 then
                return set_mario_action(m, ACT_LONG_JUMP, 0)
            end
            if m.controller.buttonPressed & B_BUTTON ~= 0 then
                return set_mario_action(m, ACT_SLIDE_KICK, 9)
            end
        end

        if vel >= 15.0 and m.pos.y <= m.waterLevel - 25.0 then
            m.pos.y = m.waterLevel
            return set_mario_action(m, ACT_ARIO_WATER_ROLL, 0)
        end

        set_mario_particle_flags(m, PARTICLE_DUST, 0)
        play_sound_with_freq_scale(SOUND_MOVING_TERRAIN_SLIDE + m.terrainSoundAddend,
            m.marioObj.header.gfx.cameraToObject, 1.0 + vel / 64.0)
    end

    if m.actionState <= 0 then
        if m.controller.buttonPressed & Y_BUTTON ~= 0 then
            play_sound_with_freq_scale(SOUND_OBJ_MRI_SHOOT, m.marioObj.header.gfx.cameraToObject, 1.25)
            audio_sample_play(SFX_STAR, m.pos, 0.5)
            m.actionState = 22
        end
    elseif m.actionState > 10 then
        local newVel = maxf(50.0, vel)
        m.slideVelX = sins(m.slideYaw) * newVel
        m.slideVelZ = coss(m.slideYaw) * newVel
        set_mario_particle_flags(m, PARTICLE_SPARKLES, 0)
    end
    if m.actionState > 0 then
        m.actionState = m.actionState - 1
    end

    local step = perform_ground_step(m)

    if step == GROUND_STEP_LEFT_GROUND then
        m.vel.y = 8.0
        return set_mario_action(m, ACT_ARIO_AIR_ROLL, 0)
    elseif step == GROUND_STEP_HIT_WALL and m.forwardVel > 5.0 then
        mario_bonk_reflection(m, 0)
        m.slideYaw = m.faceAngle.y
        m.slideVelX = sins(m.slideYaw) * vel
        m.slideVelZ = coss(m.slideYaw) * vel
    end

    local accel = max(1, math.sqrt(vel) * 0x9000)
    set_character_anim_with_accel(m, CHAR_ANIM_FORWARD_SPINNING, accel)
    smlua_anim_util_set_animation(m.marioObj, 'ario_roll')
    align_with_floor(m)

    m.actionTimer = m.actionTimer + 1
end

---@param m MarioState
local function ario_water_roll_loop(m)
    update_air_with_turn(m)

    m.pos.y = m.waterLevel - maxf(50.0 - m.forwardVel, 0.0)

    if m.forwardVel < 15.0 or m.controller.buttonDown & Z_TRIG == 0 then
        return set_mario_action(m, ACT_FREEFALL, 0)
    end

    if m.controller.buttonPressed & A_BUTTON ~= 0 then
        return set_mario_action(m, ACT_LONG_JUMP, 0)
    end
    if m.controller.buttonPressed & B_BUTTON ~= 0 then
        return set_mario_action(m, ACT_SLIDE_KICK, 9)
    end

    local step = perform_air_step(m, 0)

    if step == AIR_STEP_HIT_WALL and gServerSettings.bouncyLevelBounds == 0 then
        mario_bonk_reflection(m, 0)
    elseif step == AIR_STEP_LANDED then
        return set_mario_action(m, ACT_ARIO_ROLL, 0)
    end

    local accel = max(1, math.sqrt(m.forwardVel) * 0x9000)
    set_character_anim_with_accel(m, CHAR_ANIM_FORWARD_SPINNING, accel)
    smlua_anim_util_set_animation(m.marioObj, 'ario_roll')

    set_mario_particle_flags(m, PARTICLE_SHALLOW_WATER_WAVE | PARTICLE_WAVE_TRAIL, 0)
    play_sound_with_freq_scale(SOUND_MOVING_TERRAIN_RIDING_SHELL + m.terrainSoundAddend,
        m.marioObj.header.gfx.cameraToObject, 1.0 + m.forwardVel / 64.0)

    m.actionTimer = m.actionTimer + 1
end
local function ario_water_roll_gravity(m)
    m.vel.y = 0
end

---@param m MarioState
local function ario_air_roll_loop(m)
    update_air_with_turn(m)

    if m.controller.buttonPressed & Z_TRIG ~= 0 then
        m.actionState = 1
        play_sound(SOUND_GENERAL_SWISH_AIR_2, m.marioObj.header.gfx.cameraToObject)
    end

    -- delay bounce using state 2, helps with ground pound detection
    if m.actionState == 2 then
        m.vel.y = 45.0
        m.actionState = 0
    end
    if m.actionState == 1 then
        m.vel.y = -70.0
        set_mario_particle_flags(m, PARTICLE_DUST, 0)
    end
    local step = perform_air_step(m, 0)
    if step == AIR_STEP_LANDED then
        if m.actionState == 0 then
            if m.forwardVel < 0 then
                -- fix negative vel bug
                mario_set_forward_vel(m, 0)
            end
            m.vel.y = 0
            return set_mario_action(m, ACT_ARIO_ROLL, 0)
        elseif m.actionState == 1 then
            m.actionState = 2

            play_mario_heavy_landing_sound(m, SOUND_ACTION_TERRAIN_HEAVY_LANDING)
            set_mario_particle_flags(m, PARTICLE_MIST_CIRCLE | PARTICLE_HORIZONTAL_STAR, 0)
        end

        check_fall_damage_or_get_stuck(m, ACT_HARD_FORWARD_GROUND_KB)
        return
    elseif step == AIR_STEP_HIT_WALL and not (m.wall == nil and gServerSettings.bouncyLevelBounds ~= BOUNCY_LEVEL_BOUNDS_OFF) then
        if m.forwardVel > 10.0 then
            mario_bonk_reflection(m, 0)
        else
            m.forwardVel = 0.0
        end
    end

    local accel = max(0x10000, math.sqrt(m.forwardVel) * 0x8000)
    set_character_anim_with_accel(m, CHAR_ANIM_FORWARD_SPINNING, accel)
    smlua_anim_util_set_animation(m.marioObj, 'ario_roll')
end
local function ario_air_roll_gravity(m)
    if m.vel.y > 0 and m.controller.buttonDown & Z_TRIG == 0 then
        m.vel.y = maxf(0.0, m.vel.y - 4.0)
    end
    m.vel.y = maxf(-75.0, m.vel.y - 3.0)

    if m.pos.y < m.waterLevel - 100 then
        m.vel.y = m.vel.y * 0.75
    end
end

hook_mario_action(ACT_ARIO_ROLL, ario_roll_loop)
hook_mario_action(ACT_ARIO_WATER_ROLL, { every_frame = ario_water_roll_loop, gravity = ario_water_roll_gravity })
hook_mario_action(ACT_ARIO_AIR_ROLL, { every_frame = ario_air_roll_loop, gravity = ario_air_roll_gravity },
    INT_GROUND_POUND)


---@param m MarioState
local function ario_cap_hide_loop(m)
    if m.actionTimer == 0 then
        ario_enter_cap_model(m)
        gStateExtras[m.playerIndex].capHideHits = 0
    end
    m.hurtCounter = 0

    m.forwardVel = approach_f32_symmetric(m.forwardVel, m.intendedMag / 4.0, 1.0)
    if m.input & INPUT_NONZERO_ANALOG ~= 0 then
        m.faceAngle.y = approach_s16_symmetric(m.faceAngle.y, m.intendedYaw, 0xA00)
    end

    if m.controller.buttonDown & (X_BUTTON | Z_TRIG) ~= (X_BUTTON | Z_TRIG) then
        ario_exit_cap_model(m)
        play_sound(SOUND_OBJ2_MONTY_MOLE_APPEAR, m.marioObj.header.gfx.cameraToObject)
        return set_mario_action(m, ACT_STOP_CROUCHING, 0)
    end
    if m.controller.buttonPressed & A_BUTTON ~= 0 then
        ario_exit_cap_model(m)
        return set_mario_action(m, ACT_BACKFLIP, 0)
    end
    if m.actionState > 3 then
        ario_exit_cap_model(m)
        return set_mario_action(m, ACT_BACKWARD_GROUND_KB, 0)
    end

    mario_set_forward_vel(m, m.forwardVel)
    local step = perform_ground_step(m)

    local t = math.sin(m.actionTimer * 0.2) * m.intendedMag
    m.marioObj.header.gfx.angle.z = t * 0x40

    if step == GROUND_STEP_LEFT_GROUND then
        ario_exit_cap_model(m)
        return set_mario_action(m, ACT_FREEFALL, 0)
    end

    -- ensure that model is reset
    if check_common_object_cancels(m) ~= 0 then
        ario_exit_cap_model(m)
    end

    m.actionTimer = m.actionTimer + 1
end

---@param m MarioState
local function ario_cap_hide_water_loop(m)
    if m.actionTimer == 0 then
        ario_enter_cap_model(m)
        gStateExtras[m.playerIndex].capHideHits = 0
    end
    m.hurtCounter = 0

    m.forwardVel = approach_f32_symmetric(m.forwardVel, m.intendedMag / 4.0, 0.1)
    if m.input & INPUT_NONZERO_ANALOG ~= 0 then
        m.faceAngle.y = approach_s16_symmetric(m.faceAngle.y, m.intendedYaw, 0x200)
    end

    if m.controller.buttonDown & X_BUTTON == 0 or m.pos.y > m.waterLevel - 100 then
        ario_exit_cap_model(m)
        play_sound(SOUND_OBJ2_MONTY_MOLE_APPEAR, m.marioObj.header.gfx.cameraToObject)
        return set_mario_action(m, ACT_WATER_IDLE, 0)
    end
    if m.actionState > 3 then
        ario_exit_cap_model(m)
        return set_mario_action(m, ACT_BACKWARD_WATER_KB, 0)
    end

    mario_set_forward_vel(m, m.forwardVel)
    m.vel.y = -1.0
    if m.controller.buttonDown & A_BUTTON ~= 0 then
        m.vel.y = 3.0
    elseif m.controller.buttonDown & Z_TRIG ~= 0 then
        m.vel.y = -3.0
    end

    local step = perform_air_step(m, 0)

    local t = math.sin(m.actionTimer * 0x20)
    m.marioObj.header.gfx.pos.y = m.pos.y + t
    m.marioObj.header.gfx.angle.x = m.forwardVel * 0x200

    m.actionTimer = m.actionTimer + 1
end

hook_mario_action(ACT_ARIO_CAP_HIDE, ario_cap_hide_loop)
hook_mario_action(ACT_ARIO_CAP_HIDE_WATER, ario_cap_hide_water_loop)
