local E_MODEL_ARIO_CLOUD_PLATFORM = smlua_model_util_get_id('ario_cloud_plat_geo')
local COL_ARIO_CLOUD_PLATFORM = smlua_collision_util_get('ario_cloud_plat_collision')

local CAP_ACT_THROWN, CAP_ACT_IDLE, CAP_ACT_BOUNCED, CAP_ACT_PLATFORM, CAP_ACT_RETURN = 0, 1, 2, 3, 4

define_custom_obj_fields({oArioPlatformScale = 'f32', oArioPlatformShake = 'f32'})

---@type table<integer, fun(o:Object,m:MarioState)>
local thrownCapActs = {
    [CAP_ACT_THROWN] = function(o, m)
        object_step_without_floor_orient()

        if o.oTimer > 6 then
            if o.oBehParams == 0 then
                o.oAction = CAP_ACT_IDLE
            else
                spawn_mist_particles_with_sound(SOUND_GENERAL_VANISH_SFX)
                cur_obj_scale(0.1)
                o.oFaceAngleYaw = obj_angle_to_object(o, m.marioObj)
                o.oPosY = o.oPosY + 20
                o.oGraphYOffset = 0
                o.oAction = CAP_ACT_PLATFORM
                o.oTimer = 0
                return
            end
        end
        o.oFaceAngleYaw = o.oFaceAngleYaw + 0x1500
    end,
    [CAP_ACT_IDLE] = function(o, m)
        o.oForwardVel = approach_f32_symmetric(o.oForwardVel, 0.0, 10.0)
        object_step_without_floor_orient()

        if (o.oTimer > 32 and m.controller.buttonDown & X_BUTTON == 0) or o.oTimer > 120 then
            o.oAction = CAP_ACT_RETURN
            o.oTimer = 0
        end

        if obj_check_hitbox_overlap(o, m.marioObj) then
            set_mario_action(m, ACT_FREEFALL, 0)
            m.vel.y = 45
            spawn_mist_particles_variable(1, 0, 25.0)
            play_sound(SOUND_OBJ_MRI_SHOOT, o.header.gfx.cameraToObject)
            o.oAction = CAP_ACT_BOUNCED
            o.oTimer = 0
            return
        end
        o.oFaceAngleYaw = o.oFaceAngleYaw + 0x1500
    end,
    [CAP_ACT_BOUNCED] = function(o)
        o.header.gfx.scale.y = (o.oTimer + 1) / 10
        if o.oTimer > 5 then
            o.oAction = CAP_ACT_RETURN
            o.oTimer = 0
        end
    end,
    [CAP_ACT_PLATFORM] = function(o, m)
        if o.oTimer == 0 then
            obj_set_model_extended(o, E_MODEL_ARIO_CLOUD_PLATFORM)
            o.header.gfx.disableAutomaticShadowPos = true
            o.oForwardVel = 0
            o.oVelY = 0
        end

        load_object_collision_model()

        if m.controller.buttonPressed & X_BUTTON ~= 0 or obj_has_model_extended(m.marioObj, E_MODEL_ARIO_CLOUD) == 0 then
            spawn_mist_particles_with_sound(SOUND_OBJ2_MONTY_MOLE_APPEAR)
            obj_set_model_extended(o, CAPMODELS_ARIO_CLOUD.normal)
            cur_obj_scale(1.0)
            o.header.gfx.disableAutomaticShadowPos = false
            o.oPosY = o.oPosY - 20
            o.oGraphYOffset = 50
            o.oAction = CAP_ACT_RETURN
            o.oTimer = 0
            return
        end

        o.oArioPlatformScale = approach_f32_asymptotic(o.oArioPlatformScale, 1.0, 1.25)
        o.oArioPlatformShake = approach_f32_symmetric(o.oArioPlatformShake, 0.05, 0.01)

        local t = o.oTimer * 1092
        local xz = maxf(0.0, o.oArioPlatformScale - sins(t) * o.oArioPlatformShake)
        local y = maxf(0.0, o.oArioPlatformScale - coss(t) * o.oArioPlatformShake)
        local nearestM = nearest_mario_state_to_object(o)
        if nearestM.marioObj.platform == o then
            --squish
            y = y * 0.85
            if nearestM.action == ACT_ARIO_AIR_ROLL or nearestM.action == ACT_GROUND_POUND_LAND then
                o.oArioPlatformShake = 0.5
                o.oTimer = 1
            end
        end
        obj_scale_xyz(o, xz, y, xz)

        o.header.gfx.shadowPos.x = o.oPosX
        o.header.gfx.shadowPos.y = o.oPosY - 50
        o.header.gfx.shadowPos.z = o.oPosZ
    end,
    [CAP_ACT_RETURN] = function(o, m)
        o.oMoveAngleYaw = obj_angle_to_object(o, m.marioObj)
        o.oForwardVel = approach_f32_symmetric(o.oForwardVel, 120.0, 30.0)

        cur_obj_move_xz_using_fvel_and_yaw()

        o.oPosY = approach_f32_asymptotic(o.oPosY, m.pos.y, 0.25)

        if o.oTimer > 1 and obj_check_hitbox_overlap(o, m.marioObj) then
            play_sound(SOUND_ACTION_HANGING_STEP, o.header.gfx.cameraToObject)
            gPlayerSyncTable[m.playerIndex].capThrown = false
            -- keep players from using the platform throw again until landing
            if m.action & ACT_FLAG_AIR ~= 0 then gStateExtras[m.playerIndex].restrictCapThrow = true end
            obj_mark_for_deletion(o)
        end
        o.oFaceAngleYaw = o.oFaceAngleYaw + 0x1500
    end
}

---@param o Object
local function thrown_cap_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    o.oGraphYOffset = 50
    cur_obj_scale(1.0)
    o.oOpacity = 0xFF

    cur_obj_set_hitbox_radius_and_height(75, 50)
    o.hitboxDownOffset = -50
    cur_obj_set_hurtbox_radius_and_height(75, 50)

    o.oGravity = 0.0
    o.oDragStrength = 0.0
    o.oBuoyancy = 1.0
    o.oFriction = 1.0
    o.oBounciness = 0.0

    o.collisionData = COL_ARIO_CLOUD_PLATFORM
    o.oCollisionDistance = 1000

    network_init_object(o, true, {'oArioPlatformScale', 'oArioPlatformShake'})

    o.oArioPlatformScale = 0.01
    o.oArioPlatformShake = 0.05
end
---@param o Object
local function thrown_cap_loop(o)
    local m = gMarioStates[network_local_index_from_global(o.globalPlayerIndex)]

    thrownCapActs[o.oAction](o, m)

    if o.header.gfx.areaIndex ~= m.area.index or not gPlayerSyncTable[m.playerIndex].capThrown then
        o.header.gfx.disableAutomaticShadowPos = false
        gPlayerSyncTable[m.playerIndex].capThrown = false
        obj_mark_for_deletion(o)
    end
end

id_bhvArioThrownCap = hook_behavior(nil, OBJ_LIST_SURFACE, false, thrown_cap_init, thrown_cap_loop, 'bhvArioThrownCap')
